
#import "NSDocumentController.h"

#ifdef OPENSTEP

/* OPENSTEP doesn't have NSDocumentController, so we need a quick-n-dirty replacement. */

#import <AppKit/NSOpenPanel.h>
#import <AppKit/NSApplication.h>
#import <AppKit/NSMenuItem.h>
#import <AppKit/NSWorkspace.h>
#import "NSDocumentFrameworkPrivate.h"


static NSString *NSTypesKey             = @"NSTypes";
static NSString *NSNameKey              = @"NSName";
static NSString *NSRoleKey              = @"NSRole";
static NSString *NSHumanReadableNameKey = @"NSHumanReadableName";
static NSString *NSUnixExtensionsKey    = @"NSUnixExtensions";
static NSString *NSDOSExtensionsKey     = @"NSDOSExtensions";
static NSString *NSMacOSTypesKey        = @"NSMacOSTypes";
static NSString *NSMIMETypesKey         = @"NSMIMETypes";
static NSString *NSDocumentClassKey     = @"NSDocumentClass";

static NSString *NSDefaultOpenDirectory = @"NSDefaultOpenDirectory";

static NSString *NSEditorRole = @"Editor";
static NSString *NSViewerRole = @"Viewer";
static NSString *NSNoRole     = @"None";

#define TYPE_INFO(name) TypeInfoForName(_types, name)

static NSDictionary *TypeInfoForName(NSArray *types, NSString *typeName)
{
    int i, count = [types count];
    for (i=0; i<count;i++)
    {
        NSDictionary *dict = [types objectAtIndex:i];
        if ([[dict objectForKey:NSNameKey] isEqualToString:typeName])
            return dict;
    }

    return nil;
}

@implementation NSDocumentController

+ (void)initialize
{
}

+ (id)documentController //private
{
    return [self sharedDocumentController];
}

+ (id)allocWithZone:(NSZone *)zone
{
    return [self sharedDocumentController];
}
+ (id)sharedDocumentController
{
    static id instance = nil;

    if (instance == nil) instance = [[super allocWithZone:NULL] init];
    return instance;
}

- init
{
    NSDictionary *customDict = [[NSBundle mainBundle] infoDictionary];

    _types = [[customDict objectForKey:NSTypesKey] retain];
    _documents = [[NSMutableArray alloc] init];
    [self setShouldCreateUI:YES];

    [[[NSWorkspace sharedWorkspace] notificationCenter]
                addObserver:self
                   selector:@selector(_workspaceWillPowerOff:)
                       name:NSWorkspaceWillPowerOffNotification
                     object:nil];

    return self;
}

- (void)dealloc
{
    [[[NSWorkspace sharedWorkspace] notificationCenter] removeObserver:self];
    [_documents release];
    [_types release];
    [super dealloc];
}

- (BOOL)shouldCreateUI
{
    return _controllerFlags.shouldCreateUI;
}

- (void)setShouldCreateUI:(BOOL)flag
{
    _controllerFlags.shouldCreateUI = flag;
}

- (id)makeUntitledDocumentOfType:(NSString *)type
{
    Class documentClass = [self documentClassForType:type];
    return [[[documentClass alloc] init] autorelease];
}

- (id)makeDocumentWithContentsOfFile:(NSString *)fileName ofType:(NSString *)type
{
    Class documentClass = [self documentClassForType:type];
    return [[[documentClass alloc] initWithContentsOfFile:fileName ofType:type] autorelease];
}

/* Returns the first *editor* type, otherwise the first viewer type. */
- _defaultType
{
    static NSString *defaultType = nil;

    if (defaultType == nil)
    {
        int i, count = [_types count];

        for (i=0; i<count; i++)
        {
            NSDictionary *typeInfo = [_types objectAtIndex:i];
            NSString     *role     = [typeInfo objectForKey:NSRoleKey];
            NSString     *typeName = [typeInfo objectForKey:NSNameKey];

            if ([role isEqual:NSEditorRole]) {
                defaultType = typeName;
                break;
            }
            else if (defaultType != nil && [role isEqual:NSViewerRole]) {
                defaultType = typeName;
            }
        }
    }

    return defaultType;
}

/* These next two should really have been public. */
- (void)_addDocument:(NSDocument *)document
{
    [_documents addObject:document];
}

- (void)_removeDocument:(NSDocument *)document
{
    [_documents removeObject:document];
}

- (id)openUntitledDocumentOfType:(NSString*)type display:(BOOL)display
{
    NSDocument *document = [self makeUntitledDocumentOfType:type];

    if (document == nil) return nil;

    [self _addDocument:document];
    if ([self shouldCreateUI])
    {
        [document makeWindowControllers];
        if (display)
            [document showWindows];
    }

    return document;
}

- (id)openDocumentWithContentsOfFile:(NSString *)fileName display:(BOOL)display
{
    NSDocument *document = [self documentForFileName:fileName];

    if (document == nil)
    {
        NSString *extension = [fileName pathExtension];
        NSString *type;

        if ([extension length] == 0) extension = fileName;
        type = [self typeFromFileExtension:extension];

        if (type == nil) {      //FIXME -- localize.
            NSRunAlertPanel(@"Error", @"Can't open file '%@'.", nil, nil, nil, fileName);
            return nil;
        }

        if (document = [self makeDocumentWithContentsOfFile:fileName ofType:type])
            [self _addDocument:document];
        if ([self shouldCreateUI])
            [document makeWindowControllers];
    }

    if (display && [self shouldCreateUI])
    {
        [document showWindows];
    }

    return document;
}


/* Used to open untitled document at startup */
- (BOOL)_openUntitled
{
    NSString *type = [self _defaultType];
    if (type == nil) return NO;
    [self newDocument:self];
    return YES;
}

- _setupOpenPanel
{
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    [openPanel setDirectory:[self currentDirectory]];
    [openPanel setAllowsMultipleSelection:YES];
    return openPanel;
}

- (int)runModalOpenPanel:(NSOpenPanel *)openPanel forTypes:(NSArray *)openableFileExtensions
{
    return [openPanel runModalForTypes:openableFileExtensions];
}

- (NSArray *)_openableFileExtensions
{
    int i, count = [_types count];
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
    {
        NSDictionary *typeInfo = [_types objectAtIndex:i];
        [array addObjectsFromArray:[typeInfo objectForKey:NSUnixExtensionsKey]];
        [array addObjectsFromArray:[typeInfo objectForKey:NSDOSExtensionsKey]];
    }

    return array;
}

- (NSArray *)fileNamesFromRunningOpenPanel
{
    NSArray *types = [self _openableFileExtensions];
    NSOpenPanel *openPanel = [self _setupOpenPanel];

    if ([self runModalOpenPanel:openPanel forTypes:types])
    {
        [[NSUserDefaults standardUserDefaults]
             setObject:[openPanel directory] forKey:NSDefaultOpenDirectory];
        return [openPanel filenames];
    }

    return nil;
}

- (IBAction)saveAllDocuments:(id)sender
{
    NSDocument *document;
    NSEnumerator *docEnum = [_documents objectEnumerator];

    while (document = [docEnum nextObject])
    {
        if ([document isDocumentEdited])  //maybe we should save regardless...
        {
            [document saveDocument:sender];
        }
    }
}


- (IBAction)openDocument:(id)sender
{
    NSEnumerator *fileEnum = [[self fileNamesFromRunningOpenPanel] objectEnumerator];
    NSString *filename;

    while (filename = [fileEnum nextObject])
    {
        [self openDocumentWithContentsOfFile:filename display:YES];
    }
}

- (IBAction)newDocument:(id)sender
{
    [self openUntitledDocumentOfType:[self _defaultType] display:YES];
}


- (BOOL)closeAllDocuments
{
    NSDocument *document;
    NSEnumerator *docEnum = [_documents objectEnumerator];

    while (document = [docEnum nextObject])
    {
        if (![document canCloseDocument]) return NO;
        [document close];
        [self _removeDocument:document];
    }

    return YES;
}

- (BOOL)reviewUnsavedDocumentsWithAlertTitle:(NSString *)title cancellable:(BOOL)cancellable
{
    //FIXME -- localize.
    NSString *cancelString = (cancellable)? @"Cancel" : nil;
    int      result;

    if (![self hasEditedDocuments]) return YES;

    result = NSRunAlertPanel(title, @"You have unsaved documents.",
                             @"Review Unsaved", cancelString, @"Quit Anyways");

#define ReviewUnsaved NSAlertDefaultReturn
#define Cancel        NSAlertAlternateReturn
#define QuitAnyways   NSAlertOtherReturn

    switch (result)
    {
        case ReviewUnsaved: return [self closeAllDocuments];
        case QuitAnyways:   return YES;
        case Cancel:
        default:            return NO;
    }
}

#ifdef OPENSTEP
/*
 * App delegate methods.  Apple doesn't have these, but they put code
 * into NSApplication to call the corresponding NSDocumentController
 * methods if the app delegate doesn't implement a given delegate method.
 */
- (BOOL)application:(NSApplication *)sender openFile:(NSString *)filename
{
    return [self openDocumentWithContentsOfFile:filename display:YES] ? YES : NO;
}

- (BOOL)application:(NSApplication *)sender openTempFile:(NSString *)filename;
{
    return [self openDocumentWithContentsOfFile:filename display:YES] ? YES : NO;
}

- (BOOL)applicationOpenUntitledFile:(NSApplication *)sender
{
    return [self _openUntitled];
}

- (BOOL)application:(id)sender openFileWithoutUI:(NSString *)filename
{
    return [self openDocumentWithContentsOfFile:filename display:NO] ? YES : NO;
}

- (BOOL)applicationShouldTerminate:(NSApplication *)sender
{
    return [self reviewUnsavedDocumentsWithAlertTitle:@"Quit" cancellable:YES];
}
#endif

- (void)_workspaceWillPowerOff:(NSNotification *)notification
{
    // FIXME -- localize.
    [self reviewUnsavedDocumentsWithAlertTitle:@"Power" cancellable:NO];
}

- (NSArray *)documents
{
    return _documents;
}

- (BOOL)hasEditedDocuments;
{
    int i, count = [_documents count];

    for (i=0; i<count; i++)
    {
        if ([[_documents objectAtIndex:i] isDocumentEdited]) return YES;
    }

    return NO;
}

- (id)currentDocument
{
    return [self documentForWindow:[[NSApplication sharedApplication] mainWindow]];
}

- (NSString *)currentDirectory
{
    NSFileManager *manager = [NSFileManager defaultManager];
    NSDocument *currentDocument = [self currentDocument];
    NSString *directory = [[currentDocument fileName] stringByDeletingLastPathComponent];
    NSString *defaultsDir = [[NSUserDefaults standardUserDefaults] stringForKey:NSDefaultOpenDirectory];
    BOOL isDir = NO;

    if (directory && [manager fileExistsAtPath:directory isDirectory:&isDir] && isDir)
        return directory;
    if (defaultsDir && [manager fileExistsAtPath:defaultsDir isDirectory:&isDir] && isDir)
        return defaultsDir;

    return NSHomeDirectory();
}

- (id)documentForWindow:(NSWindow *)window
{
    int i, count = [_documents count];

    if (window == nil) return nil;

#ifdef GNUSTEP
    if ([[window windowController] isKindOfClass:[NSWindowController class]])
        return [[window windowController] document];
#else
    if ([[window delegate] isKindOfClass:[NSWindowController class]])
        return [[window delegate] document];
#endif

    for (i=0; i<count; i++)
    {
        NSDocument *document = [_documents objectAtIndex:i];
        NSArray *winControllers = [document windowControllers];
        int j, winCount = [winControllers count];

        for (j=0; j<winCount; j++)
        {
            NSWindowController *winController = [winControllers objectAtIndex:j];

            if ([winController isWindowLoaded] && [winController window] == window)
                return document;
        }
    }

    return nil;
}

- (id)documentForFileName:(NSString *)fileName
{
    int i, count = [_documents count];

    for (i=0; i<count; i++)
    {
        NSDocument *document = [_documents objectAtIndex:i];

        if ([[document fileName] isEqualToString:fileName])
            return document;
    }

    return nil;
}

- (BOOL)validateMenuItem:(NSMenuItem *)anItem
{
    if ([anItem action] == @selector(saveAllDocuments:))
        return [self hasEditedDocuments];
    return YES;
}

- (NSString *)displayNameForType:(NSString *)type
{
    NSString *name = [TYPE_INFO(type) objectForKey:NSHumanReadableNameKey];
    return name? name : type;
}

- (NSString *)typeFromFileExtension:(NSString *)fileExtension
{
    int i, count = [_types count];

    for (i=0; i<count;i++)
    {
        NSDictionary *typeInfo = [_types objectAtIndex:i];

        if ([[typeInfo objectForKey:NSUnixExtensionsKey] containsObject:fileExtension] ||
            [[typeInfo objectForKey:NSDOSExtensionsKey]  containsObject:fileExtension])
        {
            return [typeInfo objectForKey:NSNameKey];
        }
    }

    return @"";  // Apple returns blank string here
}

- (NSArray *)fileExtensionsFromType:(NSString *)type
{
    NSDictionary *typeInfo = TYPE_INFO(type);
    NSArray *unixExtensions = [typeInfo objectForKey:NSUnixExtensionsKey];
    NSArray *dosExtensions  = [typeInfo objectForKey:NSDOSExtensionsKey];

    if (!dosExtensions)  return unixExtensions;
    if (!unixExtensions) return dosExtensions;
    return [unixExtensions arrayByAddingObjectsFromArray:dosExtensions];
}

- (Class)documentClassForType:(NSString *)type
{
    NSString *className = [TYPE_INFO(type) objectForKey:NSDocumentClassKey];
    return className? NSClassFromString(className) : Nil;
}


- (NSArray *)_editorAndViewerTypesForClass:(Class)documentClass
{
    int i, count = [_types count];
    NSMutableArray *types = [NSMutableArray arrayWithCapacity:count];
    NSString *docClassName = NSStringFromClass(documentClass);

    for (i=0; i<count; i++)
    {
        NSDictionary *typeInfo  = [_types objectAtIndex:i];
        NSString     *className = [typeInfo objectForKey:NSDocumentClassKey];
        NSString     *role      = [typeInfo objectForKey:NSRoleKey];

        if ([docClassName isEqualToString:className] &&
            (role == nil || [role isEqual:NSEditorRole] || [role isEqual:NSViewerRole]))
        {
            [types addObject:[typeInfo objectForKey:NSNameKey]];
        }
    }

    return types;
}

- (NSArray *)_editorTypesForClass:(Class)documentClass
{
    int i, count = [_types count];
    NSMutableArray *types = [NSMutableArray arrayWithCapacity:count];
    NSString *docClassName = NSStringFromClass(documentClass);

    for (i=0; i<count; i++)
    {
        NSDictionary *typeInfo  = [_types objectAtIndex:i];
        NSString     *className = [typeInfo objectForKey:NSDocumentClassKey];
        NSString     *role      = [typeInfo objectForKey:NSRoleKey];

        if ([docClassName isEqualToString:className] && ([role isEqual:NSEditorRole]))
        {
            [types addObject:[typeInfo objectForKey:NSNameKey]];
        }
    }

    return types;
}

- (NSArray *)_exportableTypesForClass:(Class)documentClass
{
    // Dunno what this method is for; maybe looks for filter types
    return [self _editorTypesForClass:documentClass];
}

@end

#endif OPENSTEP

