


/* OPENSTEP doesn't have NSWindowController, so we need a quick-n-dirty replacement. */

#import "NSWindowController.h"
#ifdef OPENSTEP
#import <AppKit/NSPanel.h>
#import <AppKit/NSNibLoading.h>
#import "NSDocumentFrameworkPrivate.h"


@implementation NSWindowController

- (id)initWithWindowNibName:(NSString *)windowNibName
{
    return [self initWithWindowNibName:windowNibName owner:self];
}

- (id)initWithWindowNibName:(NSString *)windowNibName owner:(id)owner
{
    [self initWithWindow:nil];
    _windowNibName = [windowNibName retain];
    _owner = owner;
    return self;
}

- (id)initWithWindow:(NSWindow *)window
{
    [super init];

    _window = [window retain];
    _windowFrameAutosaveName = @"";
    _wcFlags.shouldCascade = YES;
    _wcFlags.shouldCloseDocument = NO;

    if (_window)
    {
        [self _windowDidLoad];
    }

    return self;
}

- (id)init
{
    return [self initWithWindowNibName:nil];
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    [_windowNibName release];
    [_windowFrameAutosaveName release];
    [_topLevelObjects release];
    [_window autorelease];
    [super dealloc];
}

- (NSString *)windowNibName
{
    return _windowNibName;
}

- (id)owner
{
    return _owner;
}

- (void)setDocument:(NSDocument *)document
{
    _document = document;
    [self _synchronizeWindowTitleWithDocumentName];
}

- (id)document
{
    return _document;
}

- (void)setWindowFrameAutosaveName:(NSString *)name
{
    [name retain];
    [_windowFrameAutosaveName release];
    _windowFrameAutosaveName = name;

    if ([self isWindowLoaded])
    {
        [[self window] setFrameAutosaveName:name? name : @""];
    }
}

- (NSString *)windowFrameAutosaveName;
{
    return _windowFrameAutosaveName;
}

- (void)setShouldCloseDocument:(BOOL)flag
{
    _wcFlags.shouldCloseDocument = flag;
}

- (BOOL)shouldCloseDocument
{
    return _wcFlags.shouldCloseDocument;
}

- (void)setShouldCascadeWindows:(BOOL)flag
{
    _wcFlags.shouldCascade = flag;
}

- (BOOL)shouldCascadeWindows
{
    return _wcFlags.shouldCascade;
}

- (void)close
{
    [_window close];
}

#ifndef GNUSTEP
// Apple's changed the NSWindow class to call NSDocument directly.  We
// can't do that, so we have to be the window delegate for the system to
// work on OPENSTEP.  Ah well.
- (BOOL)windowShouldClose:(NSWindow *)window
{
    if (_document)
        return [_document shouldCloseWindowController:self];

    return YES;
}
#endif

- (void)_windowWillClose:(NSNotification *)notification
{
    if ([notification object] == _window)
    {
        if ([_window delegate] == self) [_window setDelegate:nil];
#ifdef GNUSTEP
        if ([_window windowController] == self) [_window setWindowController:nil];
#endif

        /*
         * If the window is set to isReleasedWhenClosed, it will release
         * itself, so nil out our reference so we don't release it again.
         * We may want to unilaterally turn off the setting in the NSWindow
         * instance so it doesn't cause problems.
         * 
         * Apple's implementation doesn't seem to deal with this case, and
         * crashes if isReleaseWhenClosed is set.
         */
        if ([_window isReleasedWhenClosed])
        {
            _window = nil;
        }

        [_document _removeWindowController:self];
    }
}

- (NSWindow *)window
{
    if (_window == nil && ![self isWindowLoaded])
    {
        // Do all the notifications.  Yes, the docs say this should
        // be implemented here instead of in -loadWindow itself.
        [self windowWillLoad];
        if ([_document respondsToSelector:@selector(windowControllerWillLoadNib:)])
            [_document windowControllerWillLoadNib:self];

        [self loadWindow];
		
        /* Do this here in case someone overrides -loadWindow */
        if (_window == nil && _document)
            [self setWindow:[_document _transferWindowOwnership]];

        [self _windowDidLoad];
        if ([_document respondsToSelector:@selector(windowControllerDidLoadNib:)])
            [_document windowControllerDidLoadNib:self];
        [self _synchronizeWindowTitleWithDocumentName];
        [_window display];
    }

    return _window;
}

// Private method; the nib loading will call this.
- (void)setWindow:(NSWindow *)aWindow
{
    [_window autorelease];
    _window = [aWindow retain];
}

- (IBAction)showWindow:(id)sender
{
    NSWindow *window = [self window];

    if ([window isKindOfClass:[NSPanel class]] && [(NSPanel*)window becomesKeyOnlyIfNeeded])
        [window orderFront:sender];
    else
        [window makeKeyAndOrderFront:sender];
}

- (NSString *)windowTitleForDocumentDisplayName:(NSString *)displayName
{
    return displayName;
}

- (void)_synchronizeWindowTitleWithDocumentName
{
    if (_document && _window)
    {
        NSString *filename = [_document fileName];
        NSString *displayName = [_document _displayName];
        NSString *title = [self windowTitleForDocumentDisplayName:displayName];
        NSString *baseFile = [[filename lastPathComponent] stringByDeletingPathExtension];

        /* If they just want to display the filename, use the fancy method */
        if (filename != nil && [title isEqualToString:baseFile])
        {
            [_window setTitleWithRepresentedFilename:filename];
        }
        else
        {
            if (filename) [_window setRepresentedFilename:filename];
            [_window setTitle:title];
        }
    }
}

- (BOOL)isWindowLoaded
{
    return _wcFlags.nibIsLoaded;
}

- (void)windowDidLoad
{
}

- (void)windowWillLoad
{
}

- (void)_windowDidLoad
{
    _wcFlags.nibIsLoaded = YES;

#ifdef GNUSTEP
    [_window setWindowController:self];
#else
    // on OPENSTEP, we *have* to be the delegate -- see -windowShouldClose:
    // above
    [_window setDelegate:self];
#endif

    [[NSNotificationCenter defaultCenter]
         addObserver:self selector:@selector(_windowWillClose:)
         name:NSWindowWillCloseNotification object:_window];

    /* Make sure window sizes itself right */
    if ([_windowFrameAutosaveName length] > 0)
    {
        [_window setFrameUsingName:_windowFrameAutosaveName];
        [_window setFrameAutosaveName:_windowFrameAutosaveName];
    }

    if ([self shouldCascadeWindows])
    {
        static NSPoint nextWindowLocation  = { 0.0, 0.0 };

        /*
         * cascadeTopLeftFromPoint will "wrap" the point back to the top
         * left if the normal cascading will cause the window to go off the
         * screen. In Apple's implementation, this wraps to the extreme top
         * of the screen, and offset only a small amount from the left. The
         * first window starts at the frame as set in IB.
         */
        nextWindowLocation = [_window cascadeTopLeftFromPoint:nextWindowLocation];
    }

    [self windowDidLoad];
}

- (void)loadWindow
{
    if ([self isWindowLoaded]) return;

    if ([NSBundle loadNibNamed:_windowNibName owner:_owner])
    {
        _wcFlags.nibIsLoaded = YES;
    }
    else
    {
        NSLog(@"%@: could not load nib named %@.nib", [self class], _windowNibName);
    }
}

/*
 * There's no way I'll ever get these compatible if Apple's versions
 * actually encode anything, sigh
 */
- initWithCoder:(NSCoder *)coder
{
    return [self init];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    // What are we supposed to encode?  Window nib name?  Or should these
    // be empty, just to conform to NSCoding, so we do an -init on
    // unarchival.  ?
}

#ifndef GNUSTEP
/*
 * Evil hackery to effectively make the NSDocument part of the responder
 * chain. Apple doesn't seem to have to do this, but they may have hacked
 * the responder chain code directly to get around the problem that
 * NSDocument needs to be in the responder chain.
 */
- (BOOL)respondsToSelector:(SEL)aSel
{
    return [super respondsToSelector:aSel] || [_document respondsToSelector:aSel];
}
- (id)methodSignatureForSelector:(SEL)aSel
{
    id sig = [super methodSignatureForSelector:aSel];
    if (sig == nil) sig = [_document methodSignatureForSelector:aSel];
    return sig;
}
- (void)forwardInvocation:(NSInvocation *)invocation
{
    if ([_document respondsToSelector:[invocation selector]])
    {
        [invocation invokeWithTarget:_document];
    }
    else
    {
        [super forwardInvocation:invocation];
    }
}
#endif

@end

#endif OPENSTEP

